"""
Module contains pydantic model loaders with some platform-specific features
"""
from typing import Any, Type

from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper  # pylint: disable-msg=E0611
from pydantic.main import BaseModel  # pylint: disable-msg=E0611
from vlutils.structures.pydantic import replaceEqualSuffix

from ..errors.errors import Error
from ..errors.exception import VLException

# pydantic model root key
_ROOT_KEY = "__root__"


def getErrorMsgNLocation(exc: ValidationError) -> tuple[str, str]:
    """
    Get error message and location with `discriminated union` model names
    Args:
        exc: pydantic exception
    Returns:
        tuple with string with error location and error message
    Notes:
        - remove after https://github.com/pydantic/pydantic/issues/3947 will be resolved
    """
    locations = []

    error = exc.raw_errors[0]
    while True:
        if isinstance(error, list):
            error = error[0]
        if not isinstance(error, ErrorWrapper):
            break
        locTuple = error.loc_tuple()
        if not locTuple:
            break
        locations.append(str(locTuple[0]))
        if len(locTuple) > 1 and isinstance(locTuple[1], int):
            locations.append(str(locTuple[1]))
        if not hasattr(error.exc, "raw_errors"):
            break
        error = error.exc.raw_errors[0]

    defaultError = exc.errors()[0]
    errorMsg = defaultError["msg"]
    if defaultError["type"] == "value_error.discriminated_union.missing_discriminator":
        locations.append(replaceEqualSuffix(error.exc.discriminator_key))
        errorMsg = "field required"
    elif defaultError["type"] == "value_error.discriminated_union.invalid_discriminator":
        errorMsg = f"unexpected value; permitted: {defaultError['ctx']['allowed_values']}"
        locations.append(replaceEqualSuffix(error.exc.discriminator_key))
    if not locations:
        locations = [str(val) for val in defaultError.get("loc", [])]

    return ".".join(filter(lambda x: x != _ROOT_KEY, locations)), errorMsg


def loadDataFromJson(data: dict, model: Type[BaseModel]) -> Any:
    """
    Load data from json with pydantic
    Args:
        data: input data
        model: pydantic model

    Returns:
        initialized object
    Raises:
        VLException(Error.InvalidInputJson.format("Invalid input type."), 400, False) if invalid data type
        VLException(Error.BadInputJson.format(locationStr, errorDict["msg"]), 400, False) if any validation error
    """
    try:
        res = model(**data)
        return res
    except TypeError as exc:
        error = Error.InvalidInputJson.format("Invalid input type.")
        raise VLException(error, 400, isCriticalError=False) from exc
    except ValidationError as exc:
        raise VLException(Error.BadInputJson.format(*getErrorMsgNLocation(exc)), 400, isCriticalError=False) from exc


def loadDataFromQuery(data: dict, model: Type[BaseModel]) -> Any:
    """
    Load data from query with pydantic
    Args:
        data: input data
        model: pydantic model

    Returns:
        initialized object
    Raises:
        VLException(Error.BadQueryExtended.format(locationStr, errorDict["msg"]), 400, False) if any validation error
    """
    try:
        res = model(**data)
        return res
    except ValidationError as exc:
        errorDict = exc.errors()[0]
        if errorDict["type"] == "value_error.missing":
            raise VLException(Error.QueryParameterNotFound.format(errorDict["loc"][0]), 400, False) from exc
        locationStr = ".".join(filter(lambda x: x != _ROOT_KEY, [str(element) for element in errorDict["loc"]]))
        error = Error.BadQueryParams.format(locationStr)
        # todo add errorDict["msg"] LUNA-4591       # pylint: disable-msg=fixme
        raise VLException(error, 400, isCriticalError=False) from exc
